@@ -1,6 +1,6 @@
 /*
- * Modifications Copyright 1993, 1994, 1995, 1996, 1999,
- *  2000, 2001, 2002, 2004 by Paul Mattes.
+ * Modifications Copyright 1993, 1994, 1995, 1996, 1999, 2000, 2001, 2002,
+ *  2003, 2004, 2005, 2006 by Paul Mattes.
  * Original X11 Port Copyright 1990 by Jeff Sparkes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
@@ -85,6 +85,7 @@ static void set_formatted(void);
 static void ctlr_blanks(void);
 static Boolean  trace_primed = False;
 static unsigned char default_fg;
+static unsigned char default_bg;
 static unsigned char default_gr;
 static unsigned char default_cs;
 static unsigned char default_ic;
@@ -127,7 +128,7 @@ static unsigned char	code_table[64] = {
 	(((c1) & 0x3F) << 6) | ((c2) & 0x3F))
 
 #define ENCODE_BADDR(ptr, addr) { \
-	if ((addr) > 0xfff) { \
+    	if ((ROWS * COLS) > 0x1000) { \
 		*(ptr)++ = ((addr) >> 8) & 0x3F; \
 		*(ptr)++ = (addr) & 0xFF; \
 	} else { \
@@ -154,18 +155,21 @@ ctlr_init(unsigned cmask unused)
 void
 ctlr_reinit(unsigned cmask)
 {
+	static struct ea *real_ea_buf = NULL;
+	static struct ea *real_aea_buf = NULL;
+
 	if (cmask & MODEL_CHANGE) {
 		/* Allocate buffers */
-		if (ea_buf)
-			Free((char *)(ea_buf - 1));
-		ea_buf = (struct ea *)Calloc(sizeof(struct ea),
+		if (real_ea_buf)
+			Free((char *)real_ea_buf);
+		real_ea_buf = (struct ea *)Calloc(sizeof(struct ea),
 					     (maxROWS * maxCOLS) + 1);
-		ea_buf++;
-		if (aea_buf)
-			Free((char *)(aea_buf - 1));
-		aea_buf = (struct ea *)Calloc(sizeof(struct ea),
+		ea_buf = real_ea_buf + 1;
+		if (real_aea_buf)
+			Free((char *)real_aea_buf);
+		real_aea_buf = (struct ea *)Calloc(sizeof(struct ea),
 					      (maxROWS * maxCOLS) + 1);
-		aea_buf++;
+		aea_buf = real_aea_buf + 1;
 		Replace(zero_buf, (unsigned char *)Calloc(sizeof(struct ea),
 							  maxROWS * maxCOLS));
 		cursor_addr = 0;
@@ -312,6 +316,7 @@ ctlr_connect(Boolean ignored unused)
 	}
 
 	default_fg = 0x00;
+	default_bg = 0x00;
 	default_gr = 0x00;
 	default_cs = 0x00;
 	default_ic = 0x00;
@@ -577,13 +582,16 @@ host_cs(unsigned char cs)
 }
 
 static void
-insert_sa(int baddr, unsigned char *current_fgp, unsigned char *current_grp, unsigned char *current_csp, Boolean *anyp)
+insert_sa(int baddr, unsigned char *current_fgp, unsigned char *current_bgp,
+	unsigned char *current_grp, unsigned char *current_csp, Boolean *anyp)
 {
 	if (reply_mode != SF_SRM_CHAR)
 		return;
 
 	if (memchr((char *)crm_attr, XA_FOREGROUND, crm_nattr))
 		insert_sa1(XA_FOREGROUND, ea_buf[baddr].fg, current_fgp, anyp);
+	if (memchr((char *)crm_attr, XA_BACKGROUND, crm_nattr))
+		insert_sa1(XA_BACKGROUND, ea_buf[baddr].bg, current_bgp, anyp);
 	if (memchr((char *)crm_attr, XA_HIGHLIGHTING, crm_nattr)) {
 		unsigned char gr;
 
@@ -610,6 +618,7 @@ ctlr_read_modified(unsigned char aid_byte, Boolean all)
 	Boolean		send_data = True;
 	Boolean		short_read = False;
 	unsigned char	current_fg = 0x00;
+	unsigned char	current_bg = 0x00;
 	unsigned char	current_gr = 0x00;
 	unsigned char	current_cs = 0x00;
 
@@ -688,6 +697,7 @@ ctlr_read_modified(unsigned char aid_byte, Boolean all)
 					    ea_buf[baddr].cc) {
 						insert_sa(baddr,
 						    &current_fg,
+						    &current_bg,
 						    &current_gr,
 						    &current_cs,
 						    &any);
@@ -701,11 +711,19 @@ ctlr_read_modified(unsigned char aid_byte, Boolean all)
 						}
 						space3270out(1);
 						*obptr++ = ea_buf[baddr].cc;
-						if (!any)
-							trace_ds(" '");
-						trace_ds("%s",
-						    see_ebc(ea_buf[baddr].cc));
-						any = True;
+						if (ea_buf[baddr].cc <= 0x3f ||
+						    ea_buf[baddr].cc == 0xff) {
+							if (any)
+								trace_ds("'");
+
+							trace_ds(" %s", see_ebc(ea_buf[baddr].cc));
+							any = False;
+						} else {
+							if (!any)
+								trace_ds(" '");
+							trace_ds("%s", see_ebc(ea_buf[baddr].cc));
+							any = True;
+						}
 					}
 					INC_BA(baddr);
 				}
@@ -733,6 +751,7 @@ ctlr_read_modified(unsigned char aid_byte, Boolean all)
 			if (ea_buf[baddr].cc) {
 				insert_sa(baddr,
 				    &current_fg,
+				    &current_bg,
 				    &current_gr,
 				    &current_cs,
 				    &any);
@@ -746,10 +765,19 @@ ctlr_read_modified(unsigned char aid_byte, Boolean all)
 				}
 				space3270out(1);
 				*obptr++ = ea_buf[baddr].cc;
-				if (!any)
-					trace_ds("'");
-				trace_ds(see_ebc(ea_buf[baddr].cc));
-				any = True;
+				if (ea_buf[baddr].cc <= 0x3f ||
+				    ea_buf[baddr].cc == 0xff) {
+					if (any)
+						trace_ds("'");
+
+					trace_ds(" %s", see_ebc(ea_buf[baddr].cc));
+					any = False;
+				} else {
+					if (!any)
+						trace_ds(" '");
+					trace_ds("%s", see_ebc(ea_buf[baddr].cc));
+					any = True;
+				}
 				nbytes++;
 			}
 			INC_BA(baddr);
@@ -782,6 +810,7 @@ ctlr_read_buffer(unsigned char aid_byte)
 	Boolean		any = False;
 	int		attr_count = 0;
 	unsigned char	current_fg = 0x00;
+	unsigned char	current_bg = 0x00;
 	unsigned char	current_gr = 0x00;
 	unsigned char	current_cs = 0x00;
 
@@ -829,6 +858,14 @@ ctlr_read_buffer(unsigned char aid_byte)
 					    ea_buf[baddr].fg));
 					(*(obuf + attr_count))++;
 				}
+				if (ea_buf[baddr].bg) {
+					space3270out(2);
+					*obptr++ = XA_BACKGROUND;
+					*obptr++ = ea_buf[baddr].bg;
+					trace_ds("%s", see_efa(XA_BACKGROUND,
+					    ea_buf[baddr].bg));
+					(*(obuf + attr_count))++;
+				}
 				if (ea_buf[baddr].gr) {
 					space3270out(2);
 					*obptr++ = XA_HIGHLIGHTING;
@@ -850,6 +887,7 @@ ctlr_read_buffer(unsigned char aid_byte)
 		} else {
 			insert_sa(baddr,
 			    &current_fg,
+			    &current_bg,
 			    &current_gr,
 			    &current_cs,
 			    &any);
@@ -896,6 +934,7 @@ ctlr_snap_buffer(void)
 	register int	baddr = 0;
 	int		attr_count;
 	unsigned char	current_fg = 0x00;
+	unsigned char	current_bg = 0x00;
 	unsigned char	current_gr = 0x00;
 	unsigned char	current_cs = 0x00;
 	unsigned char   av;
@@ -918,6 +957,12 @@ ctlr_snap_buffer(void)
 				*obptr++ = ea_buf[baddr].fg;
 				(*(obuf + attr_count))++;
 			}
+			if (ea_buf[baddr].bg) {
+				space3270out(2);
+				*obptr++ = XA_BACKGROUND;
+				*obptr++ = ea_buf[baddr].fg;
+				(*(obuf + attr_count))++;
+			}
 			if (ea_buf[baddr].gr) {
 				space3270out(2);
 				*obptr++ = XA_HIGHLIGHTING;
@@ -939,6 +984,14 @@ ctlr_snap_buffer(void)
 				*obptr++ = XA_FOREGROUND;
 				*obptr++ = av;
 			}
+			av = ea_buf[baddr].bg;
+			if (current_bg != av) {
+				current_bg = av;
+				space3270out(3);
+				*obptr++ = ORDER_SA;
+				*obptr++ = XA_BACKGROUND;
+				*obptr++ = av;
+			}
 			av = ea_buf[baddr].gr;
 			if (av)
 				av |= 0xf0;
@@ -1074,6 +1127,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 	unsigned char	na;
 	int		any_fa;
 	unsigned char	efa_fg;
+	unsigned char	efa_bg;
 	unsigned char	efa_gr;
 	unsigned char	efa_cs;
 	unsigned char	efa_ic;
@@ -1099,6 +1153,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			ctlr_add_fa(buffer_addr, fa, 0); \
 			ctlr_add_cs(buffer_addr, 0); \
 			ctlr_add_fg(buffer_addr, 0); \
+			ctlr_add_bg(buffer_addr, 0); \
 			ctlr_add_gr(buffer_addr, 0); \
 			ctlr_add_ic(buffer_addr, 0); \
 			trace_ds(see_attr(fa)); \
@@ -1111,6 +1166,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 		return PDS_BAD_CMD;
 
 	default_fg = 0;
+	default_bg = 0;
 	default_gr = 0;
 	default_cs = 0;
 	default_ic = 0;
@@ -1175,6 +1231,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			cp++;		/* skip field attribute */
 			START_FIELD(*cp);
 			ctlr_add_fg(buffer_addr, 0);
+			ctlr_add_bg(buffer_addr, 0);
 			INC_BA(buffer_addr);
 			last_cmd = True;
 			last_zpt = False;
@@ -1237,6 +1294,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 					ctlr_add(buffer_addr, EBC_null, 0);
 					ctlr_add_cs(buffer_addr, 0);
 					ctlr_add_fg(buffer_addr, 0);
+					ctlr_add_bg(buffer_addr, 0);
 					ctlr_add_gr(buffer_addr, 0);
 					ctlr_add_ic(buffer_addr, 0);
 					INC_BA(buffer_addr);
@@ -1341,6 +1399,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 					ctlr_add(buffer_addr, add_c2,
 					    default_cs);
 					ctlr_add_fg(buffer_addr, default_fg);
+					ctlr_add_bg(buffer_addr, default_bg);
 					ctlr_add_gr(buffer_addr, default_gr);
 					ctlr_add_ic(buffer_addr, default_ic);
 					INC_BA(buffer_addr);
@@ -1393,6 +1452,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 				trace_ds("'");
 			ctlr_add(buffer_addr, *cp, CS_GE);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1423,12 +1483,19 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 						cp++;
 						if (appres.m3279)
 							ctlr_add_fg(buffer_addr, *cp);
+					} else if (*cp == XA_BACKGROUND) {
+						trace_ds("%s",
+						    see_efa(*cp,
+							*(cp + 1)));
+						cp++;
+						if (appres.m3279)
+							ctlr_add_bg(buffer_addr, *cp);
 					} else if (*cp == XA_HIGHLIGHTING) {
 						trace_ds("%s",
 						    see_efa(*cp,
 							*(cp + 1)));
 						cp++;
-						ctlr_add_gr(buffer_addr, *cp & 0x07);
+						ctlr_add_gr(buffer_addr, *cp & 0x0f);
 					} else if (*cp == XA_CHARSET) {
 						int cs = 0;
 
@@ -1473,6 +1540,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			na = *cp;
 			any_fa = 0;
 			efa_fg = 0;
+			efa_bg = 0;
 			efa_gr = 0;
 			efa_cs = 0;
 			efa_ic = 0;
@@ -1488,6 +1556,11 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 					cp++;
 					if (appres.m3279)
 						efa_fg = *cp;
+				} else if (*cp == XA_BACKGROUND) {
+					trace_ds("%s", see_efa(*cp, *(cp + 1)));
+					cp++;
+					if (appres.m3279)
+						efa_bg = *cp;
 				} else if (*cp == XA_HIGHLIGHTING) {
 					trace_ds("%s", see_efa(*cp, *(cp + 1)));
 					cp++;
@@ -1518,6 +1591,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 				START_FIELD(0);
 			ctlr_add_cs(buffer_addr, efa_cs);
 			ctlr_add_fg(buffer_addr, efa_fg);
+			ctlr_add_bg(buffer_addr, efa_bg);
 			ctlr_add_gr(buffer_addr, efa_gr);
 			ctlr_add_ic(buffer_addr, efa_ic);
 			INC_BA(buffer_addr);
@@ -1532,12 +1606,17 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 				trace_ds("%s", see_efa(*cp, *(cp + 1)));
 				if (appres.m3279)
 					default_fg = *(cp + 1);
+			} else if (*cp == XA_BACKGROUND)  {
+				trace_ds("%s", see_efa(*cp, *(cp + 1)));
+				if (appres.m3279)
+					default_bg = *(cp + 1);
 			} else if (*cp == XA_HIGHLIGHTING)  {
 				trace_ds("%s", see_efa(*cp, *(cp + 1)));
-				default_gr = *(cp + 1) & 0x07;
+				default_gr = *(cp + 1) & 0x0f;
 			} else if (*cp == XA_ALL)  {
 				trace_ds("%s", see_efa(*cp, *(cp + 1)));
 				default_fg = 0;
+				default_bg = 0;
 				default_gr = 0;
 				default_cs = 0;
 				default_ic = 0;
@@ -1583,6 +1662,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			}
 			ctlr_add(buffer_addr, *cp, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1606,6 +1686,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			previous = ORDER;
 			ctlr_add(buffer_addr, *cp, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1644,6 +1725,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			previous = ORDER;
 			ctlr_add(buffer_addr, *cp, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1696,14 +1778,16 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			}
 			ctlr_add(buffer_addr, add_c1, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
-			ctlr_add_gr(buffer_addr, default_ic);
+			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
 			if (add_dbcs) {
 				ctlr_add(buffer_addr, add_c2, default_cs);
 				ctlr_add_fg(buffer_addr, default_fg);
+				ctlr_add_bg(buffer_addr, default_bg);
 				ctlr_add_gr(buffer_addr, default_gr);
-				ctlr_add_gr(buffer_addr, default_ic);
+				ctlr_add_ic(buffer_addr, default_ic);
 				INC_BA(buffer_addr);
 			}
 			last_cmd = False;
@@ -1752,6 +1836,7 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 #endif /*]*/
 			ctlr_add(buffer_addr, add_c1, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1759,8 +1844,9 @@ ctlr_write(unsigned char buf[], int buflen, Boolean erase)
 			if (add_dbcs) {
 				ctlr_add(buffer_addr, add_c2, default_cs);
 				ctlr_add_fg(buffer_addr, default_fg);
+				ctlr_add_bg(buffer_addr, default_bg);
 				ctlr_add_gr(buffer_addr, default_gr);
-				ctlr_add_gr(buffer_addr, default_ic);
+				ctlr_add_ic(buffer_addr, default_ic);
 				INC_BA(buffer_addr);
 			}
 #endif /*]*/
@@ -1837,6 +1923,7 @@ ctlr_write_sscp_lu(unsigned char buf[], int buflen)
 			while ((buffer_addr / COLS) == s_row) {
 				ctlr_add(buffer_addr, EBC_null, default_cs);
 				ctlr_add_fg(buffer_addr, default_fg);
+				ctlr_add_bg(buffer_addr, default_bg);
 				ctlr_add_gr(buffer_addr, default_gr);
 				ctlr_add_ic(buffer_addr, default_ic);
 				INC_BA(buffer_addr);
@@ -1851,6 +1938,7 @@ ctlr_write_sscp_lu(unsigned char buf[], int buflen)
 			    rcba(buffer_addr), see_attr(*cp));
 			ctlr_add(buffer_addr, EBC_space, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1876,6 +1964,7 @@ ctlr_write_sscp_lu(unsigned char buf[], int buflen)
 				c = *cp;
 			ctlr_add(buffer_addr, c, CS_GE);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -1884,6 +1973,7 @@ ctlr_write_sscp_lu(unsigned char buf[], int buflen)
 		default:
 			ctlr_add(buffer_addr, *cp, default_cs);
 			ctlr_add_fg(buffer_addr, default_fg);
+			ctlr_add_bg(buffer_addr, default_bg);
 			ctlr_add_gr(buffer_addr, default_gr);
 			ctlr_add_ic(buffer_addr, default_ic);
 			INC_BA(buffer_addr);
@@ -2246,6 +2336,7 @@ ctlr_clear(Boolean can_snap)
 	unselect(0, ROWS*COLS);
 	formatted = False;
 	default_fg = 0;
+	default_bg = 0;
 	default_gr = 0;
 	default_ic = 0;
 	sscp_start = 0;
@@ -2363,7 +2454,6 @@ ctlr_add_fg(int baddr, unsigned char color)
 	}
 }
 
-#if defined(X3270_ANSI) /*[*/
 /*
  * Change the background color for a character in the 3270 buffer.
  */
@@ -2381,7 +2471,6 @@ ctlr_add_bg(int baddr, unsigned char color)
 		ea_buf[baddr].bg = color;
 	}
 }
-#endif /*]*/
 
 /*
  * Change the input control bit for a character in the 3270 buffer.
