@@ -1,6 +1,6 @@
 /*
- * Modifications Copyright 1993, 1994, 1995, 1996, 1999, 2000,
- *   2001, 2002, 2004, 2005 by Paul Mattes.
+ * Modifications Copyright 1993, 1994, 1995, 1996, 1999, 2000, 2001, 2002,
+ *   2003, 2004, 2005, 2006, 2007 by Paul Mattes.
  * Original X11 Port Copyright 1990 by Jeff Sparkes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
@@ -127,6 +127,7 @@ XrmOptionDescRec options[]= {
 #if defined(X3270_PRINTER) /*[*/
 	{ OptPrinterLu,	DotPrinterLu,	XrmoptionSepArg,	NULL },
 #endif /*]*/
+	{ OptProxy,	DotProxy,	XrmoptionSepArg,	NULL },
 	{ OptReconnect,	DotReconnect,	XrmoptionNoArg,		ResTrue },
 	{ OptSaveLines,	DotSaveLines,	XrmoptionSepArg,	NULL },
 	{ OptScripted,	DotScripted,	XrmoptionNoArg,		ResTrue },
