@@ -1,6 +1,6 @@
 /*
- * Modifications Copyright 1993, 1994, 1995, 1999,
- *  2000, 2001, 2002, 2005 by Paul Mattes.
+ * Modifications Copyright 1993, 1994, 1995, 1999, 2000, 2001, 2002, 2003,
+ *   2004, 2005, 2007 by Paul Mattes.
  * RPQNAMES modifications Copyright 2004 by Don Russell.
  * Original X11 Port Copyright 1990 by Jeff Sparkes.
  *  Permission to use, copy, modify, and distribute this software and its
@@ -30,7 +30,6 @@
 #if defined(_WIN32) /*[*/
 #include <winsock2.h>
 #include <ws2tcpip.h>
-#undef AF_INET6
 #else /*][*/
 #include <sys/socket.h>
 #include <sys/ioctl.h>
@@ -63,11 +62,14 @@
 #include "kybdc.h"
 #include "macrosc.h"
 #include "popupsc.h"
+#include "proxyc.h"
+#include "resolverc.h"
 #include "statusc.h"
 #include "tablesc.h"
 #include "telnetc.h"
 #include "trace_dsc.h"
 #include "utilc.h"
+#include "w3miscc.h"
 #include "xioc.h"
 
 #if !defined(TELOPT_NAWS) /*[*/
@@ -160,6 +162,11 @@ static char	**lus = (char **)NULL;
 static char	**curr_lu = (char **)NULL;
 static char	*try_lu = CN;
 
+static int	proxy_type = 0;
+static char	*proxy_host = CN;
+static char	*proxy_portname = CN;
+static unsigned short proxy_port = 0;
+
 static int telnet_fsm(unsigned char c);
 static void net_rawout(unsigned const char *buf, int len);
 static void check_in3270(void);
@@ -337,7 +344,8 @@ sockstart(void)
 	wVersionRequested = MAKEWORD(2, 2);
  
 	if (WSAStartup(wVersionRequested, &wsaData) != 0) {
-		fprintf(stderr, "WSAStartup failed: %ld\n", GetLastError());
+		fprintf(stderr, "WSAStartup failed: %s\n",
+			win32_strerror(GetLastError()));
 		x3270_exit(1);
 	}
  
@@ -356,30 +364,10 @@ static union {
 	struct sockaddr_in6 sin6;
 #endif /*]*/
 } haddr;
-
-static void popup_a_sockerr(char *fmt, ...) printflike(1, 2);
+socklen_t ha_len = sizeof(haddr);
 
 #if defined(_WIN32) /*[*/
-static char *
-winsock_strerror(int e)
-{
-	static char buffer[4096];
-
-	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
-	    NULL,
-	    e,
-	    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-	    buffer,
-	    sizeof(buffer),
-	    NULL) == 0) {
-	    
-	    sprintf(buffer, "Windows error %d", e);
-	}
-
-	return buffer;
-}
-
-static void
+void
 popup_a_sockerr(char *fmt, ...)
 {
 	va_list args;
@@ -388,10 +376,10 @@ popup_a_sockerr(char *fmt, ...)
 	 va_start(args, fmt);
 	 vsprintf(buffer, fmt, args);
 	 va_end(args);
-	 popup_an_error("%s: %s", buffer, winsock_strerror(socket_errno()));
+	 popup_an_error("%s: %s", buffer, win32_strerror(socket_errno()));
 }
 #else /*][*/
-static void
+void
 popup_a_sockerr(char *fmt, ...)
 {
 	va_list args;
@@ -416,16 +404,11 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 {
 	struct servent	*sp;
 	struct hostent	*hp;
-#if !defined(AF_INET6) /*[*/
-	unsigned long		lport;
-	unsigned short		port;
-	char			*ptr;
-#endif /*]*/
 	char	        	passthru_haddr[8];
 	int			passthru_len = 0;
 	unsigned short		passthru_port = 0;
-	socklen_t		ha_len = sizeof(haddr);
 	int			on = 1;
+	char			errmsg[1024];
 #if defined(OMTU) /*[*/
 	int			mtu = OMTU;
 #endif /*]*/
@@ -479,22 +462,28 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 			passthru_port = sp->s_port;
 		else
 			passthru_port = htons(3514);
-	}
+	} else if (appres.proxy != CN && !proxy_type) {
+	    	proxy_type = proxy_setup(&proxy_host, &proxy_portname);
+		if (proxy_type > 0) {
+		    	unsigned long lport;
+			char *ptr;
+			struct servent *sp;
 
-#if !defined(AF_INET6) /*[*/
-	/* get the port number */
-	lport = strtoul(portname, &ptr, 0);
-	if (ptr == portname || *ptr != '\0' || lport == 0L || lport & ~0xffff) {
-		if (!(sp = getservbyname(portname, "tcp"))) {
-			popup_an_error("Unknown port number or service: %s",
-			    portname);
-			return -1;
+			lport = strtoul(portname, &ptr, 0);
+			if (ptr == portname || *ptr != '\0' || lport == 0L ||
+				    lport & ~0xffff) {
+				if (!(sp = getservbyname(portname, "tcp"))) {
+					popup_an_error("Unknown port number "
+						"or service: %s", portname);
+					return -1;
+				}
+				current_port = ntohs(sp->s_port);
+			} else
+				current_port = (unsigned short)lport;
 		}
-		port = sp->s_port;
-	} else
-		port = htons((unsigned short)lport);
-	current_port = ntohs(port);
-#endif /*]*/
+		if (proxy_type < 0)
+		    	return -1;
+	}
 
 	/* fill in the socket address of the given host */
 	(void) memset((char *) &haddr, 0, sizeof(haddr));
@@ -504,68 +493,28 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 			       passthru_len);
 		haddr.sin.sin_port = passthru_port;
 		ha_len = sizeof(struct sockaddr_in);
+	} else if (proxy_type > 0) {
+	    	if (resolve_host_and_port(proxy_host, proxy_portname,
+			    &proxy_port, &haddr.sa, &ha_len, errmsg,
+			    sizeof(errmsg)) < 0) {
+		    	popup_an_error(errmsg);
+		    	return -1;
+		}
 	} else {
 #if defined(LOCAL_PROCESS) /*[*/
 		if (ls) {
 			local_process = True;
 		} else {
 #endif /*]*/
-#if defined(AF_INET6) /*[*/
-			struct addrinfo hints, *res;
-			int rc;
-
-			/* Use getaddrinfo(). */
 #if defined(LOCAL_PROCESS) /*[*/
 			local_process = False;
 #endif /*]*/
-			(void) memset(&hints, '\0', sizeof(struct addrinfo));
-			hints.ai_flags = 0;
-			hints.ai_family = PF_UNSPEC;
-			hints.ai_socktype = SOCK_STREAM;
-			hints.ai_protocol = IPPROTO_TCP;
-			rc = getaddrinfo(host, portname, &hints, &res);
-			if (rc != 0) {
-				popup_an_error("%s/%s: %s", hostname, portname,
-						gai_strerror(rc));
-				return -1;
+			if (resolve_host_and_port(host, portname,
+				    &current_port, &haddr.sa, &ha_len,
+				    errmsg, sizeof(errmsg)) < 0) {
+			    	popup_an_error(errmsg);
+			    	return -1;
 			}
-			switch (res->ai_family) {
-			case AF_INET:
-				current_port = ntohs(((struct sockaddr_in *)res->ai_addr)->sin_port);
-				break;
-			case AF_INET6:
-				current_port = ntohs(((struct sockaddr_in6 *)res->ai_addr)->sin6_port);
-				break;
-			default:
-				popup_an_error("%s: unknown family %d",
-						hostname, res->ai_family);
-				freeaddrinfo(res);
-				return -1;
-			}
-			(void) memcpy(&haddr.sa, res->ai_addr, res->ai_addrlen);
-			ha_len = res->ai_addrlen;
-			freeaddrinfo(res);
-
-#else /*][*/
-			/* Use gethostbyname(). */
-			hp = gethostbyname(host);
-			if (hp == (struct hostent *) 0) {
-				haddr.sin.sin_family = AF_INET;
-				haddr.sin.sin_addr.s_addr = inet_addr(host);
-				if (haddr.sin.sin_addr.s_addr == (unsigned long)-1) {
-					popup_an_error("Unknown host:\n%s",
-					    hostname);
-					return -1;
-				}
-			}
-			else {
-				haddr.sin.sin_family = hp->h_addrtype;
-				(void) memmove(&haddr.sin.sin_addr, hp->h_addr,
-						   hp->h_length);
-			}
-			haddr.sin.sin_port = port;
-			ha_len = sizeof(struct sockaddr_in);
-#endif /*]*/
 #if defined(LOCAL_PROCESS) /*[*/
 		}
 #endif /*]*/
@@ -699,15 +648,15 @@ net_connect(const char *host, char *portname, Boolean ls, Boolean *resolving,
 
 		sock_handle = CreateEvent(NULL, TRUE, FALSE, ename);
 		if (sock_handle == NULL) {
-			fprintf(stderr, "Cannot create socket handle: %ld\n",
-			    GetLastError());
+			fprintf(stderr, "Cannot create socket handle: %s\n",
+			    win32_strerror(GetLastError()));
 			x3270_exit(1);
 		}
 	}
 	if (WSAEventSelect(sock, sock_handle, FD_READ | FD_CONNECT | FD_CLOSE)
 		    != 0) {
-		fprintf(stderr, "WSAEventSelect failed: %ld\n",
-		    GetLastError());
+		fprintf(stderr, "WSAEventSelect failed: %s\n",
+		    win32_strerror(GetLastError()));
 		x3270_exit(1);
 	}
 
@@ -774,6 +723,19 @@ setup_lus(void)
 static void
 net_connected(void)
 {
+	if (proxy_type > 0) {
+
+		/* Negotiate with the proxy. */
+	    	trace_dsn("Connected to proxy server %s, port %u.\n",
+			proxy_host, proxy_port);
+
+	    	if (proxy_negotiate(proxy_type, sock, hostname,
+			    current_port) < 0) {
+		    	host_disconnect(True);
+			return;
+		}
+	}
+
 	trace_dsn("Connected to %s, port %u%s.\n", hostname, current_port,
 	    ssl_host? " via SSL": "");
 
@@ -953,8 +915,8 @@ net_input(void)
 					return;
 				default:
 					fprintf(stderr,
-					    "second connect() failed: %d\n",
-					    err);
+					    "second connect() failed: %s\n",
+					    win32_strerror(err));
 					x3270_exit(1);
 				}
 			}
@@ -975,7 +937,12 @@ net_input(void)
 		else
 #else /*][*/
 #endif /*]*/
-		nr = recv(sock, (char *) netrbuf, BUFSZ, 0);
+#if defined(LOCAL_PROCESS) /*[*/
+		if (local_process)
+		    	nr = read(sock, (char *) netrbuf, BUFSZ);
+		else
+#endif /*]*/
+			nr = recv(sock, (char *) netrbuf, BUFSZ, 0);
 		if (nr < 0) {
 			if (socket_errno() == SE_EWOULDBLOCK) {
 				return;
@@ -1509,7 +1476,7 @@ tn3270e_request(void)
 
 	trace_dsn("SENT %s %s DEVICE-TYPE REQUEST %.*s%s%s "
 		   "%s\n",
-	    cmd(SB), opt(TELOPT_TN3270E), strlen(termtype), tt_out + 5,
+	    cmd(SB), opt(TELOPT_TN3270E), (int)strlen(termtype), tt_out + 5,
 	    (try_lu != CN && *try_lu) ? " CONNECT " : "",
 	    (try_lu != CN && *try_lu) ? try_lu : "",
 	    cmd(SE));
@@ -1988,7 +1955,12 @@ net_rawout(unsigned const char *buf, int len)
 			nw = SSL_write(ssl_con, (const char *) buf, n2w);
 		else
 #endif /*]*/
-		nw = send(sock, (const char *) buf, n2w, 0);
+#if defined(LOCAL_PROCESS) /*[*/
+		if (local_process)
+			nw = write(sock, (const char *) buf, n2w);
+		else
+#endif /*]*/
+			nw = send(sock, (const char *) buf, n2w, 0);
 		if (nw < 0) {
 #if defined(HAVE_LIBSSL) /*[*/
 			if (ssl_con != NULL) {
@@ -3245,7 +3217,7 @@ client_info_callback(INFO_CONST SSL *s, int where, int ret)
 			    SSL_state_string_long(s));
 		} else if (ret < 0) {
 			unsigned long e;
-			char err_buf[121];
+			char err_buf[1024];
 
 			err_buf[0] = '\n';
 			e = ERR_get_error();
@@ -3253,8 +3225,8 @@ client_info_callback(INFO_CONST SSL *s, int where, int ret)
 				(void) ERR_error_string(e, err_buf + 1);
 #if defined(_WIN32) /*[*/
 			else if (GetLastError() != 0)
-				sprintf(err_buf + 1, "error %ld",
-				    GetLastError());
+				strcpy(err_buf + 1,
+					win32_strerror(GetLastError()));
 #else /*][*/
 			else if (errno != 0)
 				strcpy(err_buf + 1, strerror(errno));
@@ -3443,3 +3415,33 @@ net_getsockname(void *buf, int *len)
 		return -1;
 	return getsockname(sock, buf, (socklen_t *)(void *)len);
 }
+
+/* Return a text version of the current proxy type, or NULL. */
+char *
+net_proxy_type(void)
+{
+    	if (proxy_type > 0)
+	    	return proxy_type_name(proxy_type);
+	else
+	    	return NULL;
+}
+
+/* Return the current proxy host, or NULL. */
+char *
+net_proxy_host(void)
+{
+    	if (proxy_type > 0)
+	    	return proxy_host;
+	else
+	    	return NULL;
+}
+
+/* Return the current proxy port, or NULL. */
+char *
+net_proxy_port(void)
+{
+    	if (proxy_type > 0)
+	    	return proxy_portname;
+	else
+	    	return NULL;
+}
