@@ -1,5 +1,6 @@
 /*
- * Copyright 1993, 1994, 1995, 1999, 2000, 2001, 2002, 2004 by Paul Mattes.
+ * Copyright 1993, 1994, 1995, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2007
+ *   by Paul Mattes.
  *  Permission to use, copy, modify, and distribute this software and its
  *  documentation for any purpose and without fee is hereby granted,
  *  provided that the above copyright notice appear in all copies and that
@@ -26,6 +27,9 @@
 #include <X11/StringDefs.h>
 #include <X11/Xaw/Dialog.h>
 #endif /*]*/
+#if defined(_WIN32) /*[*/
+#include <windows.h>
+#endif /*]*/
 #include <errno.h>
 #include <signal.h>
 #include <stdarg.h>
@@ -47,6 +51,7 @@
 #include "telnetc.h"
 #include "trace_dsc.h"
 #include "utilc.h"
+#include "w3miscc.h"
 
 /* Maximum size of a tracefile header. */
 #define MAX_HEADER_SIZE		(10*1024)
@@ -65,6 +70,8 @@
 static int      dscnt = 0;
 #if !defined(_WIN32) /*[*/
 static int      tracewindow_pid = -1;
+#else /*][*/
+static HANDLE	tracewindow_handle = NULL;
 #endif /*]*/
 static FILE    *tracef = NULL;
 static FILE    *tracef_pipe = NULL;
@@ -213,13 +220,19 @@ vwtrace(const char *fmt, va_list args)
 	if (tracef_bufptr != CN) {
 		tracef_bufptr += vsprintf(tracef_bufptr, fmt, args);
 	} else if (tracef != NULL) {
-		int nw;
+		int n2w, nw;
+		char buf[16384];
 
-		nw = vfprintf(tracef, fmt, args);
-		if (nw > 0) {
+		buf[0] = 0;
+		(void) vsnprintf(buf, sizeof(buf), fmt, args);
+		buf[sizeof(buf) - 1] = '\0';
+		n2w = strlen(buf);
+
+		nw = fwrite(buf, n2w, 1, tracef);
+		if (nw == 1) {
 			tracef_size += nw;
 			fflush(tracef);
-		} else if (nw < 0) {
+		} else {
 			if (errno != EPIPE
 #if defined(EILSEQ) /*[*/
 					   && errno != EILSEQ
@@ -233,10 +246,12 @@ vwtrace(const char *fmt, va_list args)
 				stop_tracing();
 		}
 		if (tracef_pipe != NULL) {
-			nw = vfprintf(tracef_pipe, fmt, args);
-			if (nw < 0) {
+			nw = fwrite(buf, n2w, 1, tracef_pipe);
+			if (nw != 1) {
 				(void) fclose(tracef_pipe);
 				tracef_pipe = NULL;
+			} else {
+			    	fflush(tracef_pipe);
 			}
 		}
 	}
@@ -406,11 +421,12 @@ create_tracefile_header(const char *mode)
 	save_yourself();
 	wtrace(" Command: %s\n", command_string);
 	wtrace(" Model %s", model_name);
+#if defined(X3270_DISPLAY) || (defined(C3270) && !defined(_WIN32)) /*[*/
 	wtrace(", %s display", appres.mono ? "monochrome" : "color");
+#endif /*]*/
 	if (appres.extended)
 		wtrace(", extended data stream");
-	if (!appres.mono)
-		wtrace(", %scolor", appres.m3279 ? "full " : "pseudo-");
+	wtrace(", %s emulation", appres.m3279 ? "color" : "monochrome");
 	wtrace(", %s charset", get_charset_name());
 	if (appres.apl_mode)
 		wtrace(", APL mode");
@@ -684,6 +700,45 @@ tracefile_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 	}
 #endif /*]*/
 
+#if defined(_WIN32) /*[*/
+	/* Start the monitor window. */
+	if (tracef != stdout && appres.trace_monitor) {
+		STARTUPINFO startupinfo;
+		PROCESS_INFORMATION process_information;
+		char *path;
+		char *args;
+
+	    	(void) memset(&startupinfo, '\0', sizeof(STARTUPINFO));
+		startupinfo.cb = sizeof(STARTUPINFO);
+		startupinfo.lpTitle = tfn;
+		(void) memset(&process_information, '\0',
+			      sizeof(PROCESS_INFORMATION));
+		path = xs_buffer("%scatf.exe", instdir);
+		args = xs_buffer("\"%scatf.exe\" \"%s\"", instdir, tfn);
+		if (CreateProcess(
+		    path,
+		    args,
+		    NULL,
+		    NULL,
+		    FALSE,
+		    CREATE_NEW_CONSOLE,
+		    NULL,
+		    NULL,
+		    &startupinfo,
+		    &process_information) == 0) {
+		    	popup_an_error("CreateProcess(%s) failed: %s",
+				path, win32_strerror(GetLastError()));
+			Free(path);
+			Free(args);
+		} else {
+			Free(path);
+		    	Free(args);
+			tracewindow_handle = process_information.hProcess;
+			CloseHandle(process_information.hThread);
+		}
+	}
+#endif /*]*/
+
 	Free(tfn);
 
 	/* We're really tracing, turn the flag on. */
@@ -718,7 +773,7 @@ no_tracefile_callback(Widget w, XtPointer client_data,
 static void
 tracefile_on(int reason, enum toggle_type tt)
 {
-	char tracefile_buf[256];
+	char *tracefile_buf = NULL;
 	char *tracefile;
 
 	if (tracef != (FILE *)NULL)
@@ -733,9 +788,10 @@ tracefile_on(int reason, enum toggle_type tt)
 		tracefile = appres.trace_file;
 	else {
 #if defined(_WIN32) /*[*/
-		(void) sprintf(tracefile_buf, "x3trc.%d.txt", getpid());
+		tracefile_buf = xs_buffer("%sx3trc.%u.txt", myappdata,
+			getpid());
 #else /*][*/
-		(void) sprintf(tracefile_buf, "%s/x3trc.%d", appres.trace_dir,
+		tracefile_buf = xs_buffer("%s/x3trc.%u", appres.trace_dir,
 			getpid());
 #endif /*]*/
 		tracefile = tracefile_buf;
@@ -746,6 +802,8 @@ tracefile_on(int reason, enum toggle_type tt)
 #endif /*]*/
 	{
 		tracefile_callback((Widget)NULL, tracefile, PN);
+		if (tracefile_buf != NULL)
+		    	Free(tracefile_buf);
 		return;
 	}
 #if defined(X3270_DISPLAY) /*[*/
@@ -765,6 +823,9 @@ tracefile_on(int reason, enum toggle_type tt)
 
 	popup_popup(trace_shell, XtGrabExclusive);
 #endif /*]*/
+
+	if (tracefile_buf != NULL)
+		Free(tracefile_buf);
 }
 
 /* Close the trace file. */
@@ -779,6 +840,13 @@ tracefile_off(void)
 	if (tracewindow_pid != -1)
 		(void) kill(tracewindow_pid, SIGKILL);
 	tracewindow_pid = -1;
+#else /*][*/
+	if (tracewindow_handle != NULL) {
+	    	TerminateProcess(tracewindow_handle, 0);
+		CloseHandle(tracewindow_handle);
+		tracewindow_handle = NULL;
+	}
+
 #endif /*]*/
 	stop_tracing();
 }
@@ -944,19 +1012,26 @@ onescreen_callback(Widget w, XtPointer client_data, XtPointer call_data unused)
 void
 toggle_screenTrace(struct toggle *t unused, enum toggle_type tt)
 {
-	char tracefile_buf[256];
+	char *tracefile_buf = NULL;
 	char *tracefile;
 
 	if (toggled(SCREEN_TRACE)) {
 		if (appres.screentrace_file)
 			tracefile = appres.screentrace_file;
 		else {
-			(void) sprintf(tracefile_buf, "%s/x3scr.%d",
+#if defined(_WIN32) /*[*/
+			tracefile_buf = xs_buffer("%sx3scr.%u.txt",
+				myappdata, getpid());
+#else /*][*/
+			tracefile_buf = xs_buffer("%s/x3scr.%u",
 				appres.trace_dir, getpid());
+#endif /*]*/
 			tracefile = tracefile_buf;
 		}
 		if (tt == TT_INITIAL || tt == TT_ACTION) {
 			(void) screentrace_cb(NewString(tracefile));
+			if (tracefile_buf != NULL)
+				Free(tracefile_buf);
 			return;
 		}
 #if defined(X3270_DISPLAY) /*[*/
@@ -978,6 +1053,9 @@ toggle_screenTrace(struct toggle *t unused, enum toggle_type tt)
 			do_screentrace();
 		(void) fclose(screentracef);
 	}
+
+	if (tracefile_buf != NULL)
+		Free(tracefile_buf);
 }
 
 #endif /*]*/
